/****************************************************************************/
/* Mar/2007                                                                 */
/* Peter Boyle.                                                             */
/*                                                                          */
/* bfm.h                                                                   */
/* C++ header file DWF & Wilson CG inverter with SIMD support               */
/*                                                                          */
/****************************************************************************/

#ifndef INCLUDED_DWF_PAB_H
#define INCLUDED_DWF_PAB_H

#undef PLL_HACK
#define BAGEL_THREADS (64)
#include <math.h>
#include <complex>
//#include <mcheck.h>

#include <string.h>
#include <bfm_options.h>
#include <bagel_int.h>
#include <stdlib.h>
#include <vector>

#ifdef BFM_QDP_BINDING
#include <qdp.h>
#define CHROMA_3
#endif

#include <zolotarev.h>

#define BFM_REPRO


#define BFM_BZERO
#ifdef BFM_BZERO
#define bzero(A,B) memset(A,0,B);
#endif

#undef LIST_ENGINE
#ifdef LIST_ENGINE
//#include <spi/include/l1p/pprefetch.h>
#include <pprefetch.h>
#endif

typedef double QDPdouble ;

enum BfmMemType { mem_fast, mem_sendbuf, mem_slow };

enum BfmSolver { 

                 DWF, 
 		 DWFrb4d, // Chroma style preconditioning

		 WilsonFermion, 
		 // Clover, // Unimplemented
		 WilsonTM,  
		 WilsonNN, // Next Nearest neighbour Wilson term (Eguchi-Kawamoto support)

		 ////////////////////////////////////
		 // Group of generalised 5d actions
		 // Kerne/Representation/approximation Combos
		 ////////////////////////////////////
		 HwPartFracZolo, // KEK's approach
		 HwContFracZolo, // Edwards, Kennedy et al prefer this
		 HwPartFracTanh, // Never been looked at!
		 HwContFracTanh, // Never been looked at!
		 HwCayleyZolo, // Chiu Optimal
		 HtCayleyZolo, // Chiu Improved
		 HwCayleyTanh, // Scaled shamir
		 HmCayleyTanh, // Scaled shamir
		 HtCayleyTanh, // Plain old DWF.
		 DWFTransfer,    // DWF transfer matrix for testing purposes
		 DWFTransferInv,  // DWF transfer inverse matrix for testing purposes
		 HtContFracTanh,
		 HtContFracZolo


};

typedef void * Fermion_t;
typedef Fermion_t bfm_fermion [2];

class bfmarg {

public:
 

   BfmSolver solver;
   int solveMobiusDminus;
   int Ls;            /*Fifth dimension - set to 1 for wilson/clover */
   double M5;         /*Five-D mass- dont care for wilson/clover */
   double mass;       /*fermion mass*/
   double twistedmass;/*twisted fermion mass ( +/-ig5 mu)*/
   double Csw;        /*Clover coefficient - unimplemented as yet*/

   int node_latt[4];  /*Node subvolume*/
   int local_comm[4]; /*are comms periodic/trivial in this axis      */
   int ncoor[4];      /*location of node in 4d node grid*/

   double zolo_delta; // minimax error
   double zolo_lo;    // For overlap Zolo
   double zolo_hi;

   int list_engine;
   int list_length;
   int powerloop;
   int time_report_iter;
   int max_iter;      // CG control
   double residual;   // CG control
   int precon_5d;     // Boolean selection of 4d or 5d checkerboarding in DWF solves.
                      // Set to 0 for Wilson solves.
   static char * watchfile;
   static int reproduce;
   static int reproduce_checksum;
   static int reproduce_master_check;
   static int threads;
   static int verbose;
   static int part_frac_chroma_convention;
   static int onepluskappanorm;
   static double mobius_scale;
   static int CGdiagonalMee;
   
   static void UseCGdiagonalMee(int useit){ CGdiagonalMee = useit;};
   static void WatchFile(char * file) {watchfile=file;};
   static void ReproduceChecksum(int rp) { reproduce_checksum = rp ;};
   static void ReproduceMasterCheck(int rp) { reproduce_master_check = rp ;};
   static void Reproduce(int r) { reproduce = r ;};
   static void Verbose(int v) { verbose = v ;};
   static void Threads  (int t) { threads =t;}
   static void OnePlusKappaNorm  (int yesno) { onepluskappanorm = yesno ; }


   bfmarg(){
     list_engine=0;
     list_length=0;
     time_report_iter = 10000000;
     twistedmass=0;
     Csw=0;
     solveMobiusDminus=0;
     powerloop=0;
     max_iter=5000;
     residual = 1.0e-7;
     precon_5d = 0;
     ncoor[0]=0;
     ncoor[1]=0;
     ncoor[2]=0;
     ncoor[3]=0;
   };

};



/* DWF matrix */

// Internally the code works only with a logical/data-parallel subnodes.
// Interaction between logical subnodes only happens during comms
// This class operates on "internal" layout vectors
// The CG routine will support import/export to more standard orderings.
  /* Constant definitions */
  enum {
    ND  = 4,
    SPINOR_SIZE = 24,
    HALF_SPINOR_SIZE = 12,
    GAUGE_SIZE = 72,
    Nmu = 4,
    Ncb = 2,
    NMinusPlus = 2,
    Minus=0,
    Plus =1,
    DaggerYes=1,
    DaggerNo =0,
    SingleToDouble=1,
    DoubleToSingle=0,
    Odd=1,
    Even=0
  };


#include <bfm_thread.h>

template <class Float>
class bfmbase : public bfmarg , public ThreadModel {

 protected:

 public:
  int nsimd;

  int iter;
  int ncb;
  int node_cbvol;
  int cbLs;
  int simd_latt[ND];
  int simd_grid[ND];
  integer simd_cbvol;
  int simd_nbound[ND];
  int simd_allbound;
  int base_parity;
  int halo_depth; // Use to support next-nearest neighbour Hamber-Wu-Eguchi-Kawamoto


  Float *u;

  /////////////////////////////////////////////////////////////////
  // GeneralisedFiveDim
  /////////////////////////////////////////////////////////////////

  Fermion_t g5d_tmp;
  void GeneralisedFiveDimEnd(void);
  void GeneralisedFiveDimInit(void);
  int IsGeneralisedFiveDim(void){
    if( solver == HwPartFracZolo   ) return 1;
    if( solver == HwContFracZolo   ) return 1;
    if( solver == HwPartFracTanh   ) return 1;
    if( solver == HwContFracTanh   ) return 1;
    if( solver == HwCayleyTanh      ) return 1;
    if( solver == HtCayleyTanh      ) return 1;
    if( solver == HmCayleyTanh      ) return 1;
    if( solver == HwCayleyZolo      ) return 1;
    if( solver == HtCayleyZolo      ) return 1;
    if( solver == DWFTransfer       ) return 1;
    if( solver == DWFTransferInv    ) return 1;
    return 0;
  }
  char *SolverString( BfmSolver solver )
  {
    if(solver ==DWF)           return (char *) "DWF";
    if(solver ==DWFrb4d)       return (char *) "DWFrb4d";
    if(solver ==WilsonFermion) return (char *) "WilsonFermion";
    if(solver ==WilsonTM)      return (char *) "WilsonTM";
    if(solver ==WilsonNN)      return (char *) "WilsonNN";
    ////////////////////////////////////
    // Group of generalised 5d actions
    // Kerne/Representation/approximation Combos
    ////////////////////////////////////
    if(solver == HwPartFracZolo) return (char *) "HwPartFracZolo";
    if(solver == HwContFracZolo) return (char *) "HwContFracZolo";
    if(solver == HwPartFracTanh) return (char *) "HwPartFracTanh";
    if(solver == HwContFracTanh) return (char *) "HwContFracTanh";
    if(solver == HwCayleyZolo) return (char *) "HwCayleyZolo";
    if(solver == HtCayleyZolo) return (char *) "HtCayleyZolo";
    if(solver == HwCayleyTanh) return (char *) "HwCayleyTanh";
    if(solver == HmCayleyTanh) return (char *) "HmCayleyTanh";
    if(solver == HtCayleyTanh) return (char *) "HtCayleyTanh";
    if(solver == DWFTransfer) return (char *) "DWFTransfer";
    if(solver == DWFTransferInv) return (char *) "DWFTransferInv";
    printf("Unknown solver type");
    exit(-1);
  };

  private:
  double eps;
  public:
  zolotarev_data *zdata;

  // Part frac
  std::vector<double> p; 
  std::vector<double> q;

  // Cont frac
  std::vector<double> Beta;
  std::vector<double> cc;;
  std::vector<double> sqrt_cc;

  // Cayley form Moebius (tanh and zolotarev)
  std::vector<double> omega; 
  std::vector<double> bs;    // S dependent coeffs
  std::vector<double> cs;    
  std::vector<double> as;    

  // For preconditioning Cayley form
  std::vector<double> bee;    
  std::vector<double> cee;    
  std::vector<double> beo;    
  std::vector<double> ceo;    
  // LDU factorisation of the eeoo matrix
  std::vector<double> lee;    
  std::vector<double> leem;    
  std::vector<double> uee;    
  std::vector<double> ueem;    
  std::vector<double> dee;    

  void G5D_gnuplot_approx(void); 

  void G5D_DW(Fermion_t psi[2],Fermion_t chi,int cb,int dag,double scale=1.0); 
  void G5D_HW(Fermion_t psi[2],Fermion_t chi,int cb,int dag); 
  void G5D_Dminus(Fermion_t psi[2],Fermion_t chi[2],int dag);
  void G5D_Dplus(Fermion_t psi[2],Fermion_t chi[2],int dag);

  void G5D_ConsvCurrents(Fermion_t sol[2]);

  void G5D_Transfer   (Fermion_t psi[2],Fermion_t chi[2]); 
  void G5D_TransferNumDen(Fermion_t psi[2],Fermion_t chi[2],int sgn,int dag);  

  void G5D_Munprec(Fermion_t psi[2], 
		   Fermion_t chi[2], 
		   int dag);

  double G5D_Mprec(Fermion_t chi, 
		   Fermion_t psi, 
		   Fermion_t tmp, 
		   int dag,int donrm=0) ;


  void G5D_Munprec_Cayley(Fermion_t psi[2], 
			    Fermion_t chi[2], 
			    int dag);

  void G5D_Munprec_PartFrac(Fermion_t psi[2], 
				  Fermion_t chi[2], 
				  int dag);
  void G5D_Munprec_ContFrac(Fermion_t psi[2], 
			    Fermion_t chi[2], 
			    int dag);

  // Preconditioned support
  void G5D_Meo(Fermion_t chi, 
	       Fermion_t psi,
	       int result_cb,
	       int dag);

  void G5D_Mooee(Fermion_t chi, 
		 Fermion_t psi, 
		 int dag) ;

  void G5D_MooeeInv(Fermion_t chi, 
		    Fermion_t psi, 
		    int dag) ;

  // Optimised version
  void G5D_CayleyMooeeFiveD(Fermion_t psi,
		       Fermion_t chi,
		       int dag);

  void DperpDWFcompat_cpp (Fermion_t psi,Fermion_t chi,int cb,int dag);
  void axpby_ssp_internal(Fermion_t out,double al, double ah,Fermion_t x,
			  double bl, double bh,Fermion_t y,int sxo,int sy);
  void axpby_ssp     (Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy);
  void axpby_ssp_proj(Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy,int psign);
  void ag5xpby_ssp_proj(Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy,int psign);

  void ag5xpby_ssp   (Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy);
  void ag5xpbg5y_ssp (Fermion_t out,double a,Fermion_t x,double b,Fermion_t y,int sxo,int sy);

  //////////////////////////////////////////////////
  // Reproducibility checking
  //////////////////////////////////////////////////
  enum { ReproIdle, ReproRecord , ReproCheck };

#ifdef BFM_REPRO
  int repro_entry;
  int csum_entry;
  void repro_control(int state) ;
  void repro(double d,int who) ;
  void repro(uint64_t sum) ;
  std::vector< std::vector<double> > repro_log;
  std::vector<uint64_t> csum_log;
#else
  void repro_control(int state) {};
  void repro(double d, int who) {};
  void repro(uint64_t) {};
#endif

  int repro_state;
  // End hack
/*
 * Build up enough infrastructure to write a generic solver
 * with abstracted internal structure.
 *
 * With good luck and a following wind, I should be able to cover 
 * Wilson/Clover/TwistedMass/DWF in one go.
 *
 * I will explicitly implement CG below for the congenitally lazy.
 * 
 * I would love to write an equivalent "C" output for SSE3.
 *
 */


 public:

/******************************************************************************************************
 * Public Methods that are common to all implementations are in the base class
 ******************************************************************************************************/


  /**************************************************************************
   * This is the QDP visible glue
   **************************************************************************
   */
#ifdef BFM_QDP_BINDING
  virtual void   importGauge(multi1d<LatticeColorMatrix> &u);
  virtual void   importGaugeNN(multi1d<LatticeColorMatrix> &u);//next nearest
  void           importFermion  (LatticeFermion &psi, Fermion_t handle,int cb,int s=0);
  void           exportFermion  (LatticeFermion &psi, Fermion_t handle,int cb,int s=0);
  // Handles DWF and G5D surface terms
  void           importPhysicalFermion(LatticeFermion &psi, Fermion_t handle,int cb);
  void           exportPhysicalFermion(LatticeFermion &psi, Fermion_t handle,int cb);
  // for DWF
  void      importFermion  (multi1d<LatticeFermion> &psi, Fermion_t handle,int cb);
  void      exportFermion  (multi1d<LatticeFermion> &psi, Fermion_t handle,int cb);
#endif

  /**************************************************************************
   * Generic interface for use in creating external solvers
   **************************************************************************
   */
  int CheckStopWatchFile(void);
  Fermion_t allocFermion   (int mem_type=mem_slow); 
  void      freeFermion    (Fermion_t handle);
  Fermion_t threadedAllocFermion   (int mem_type=mem_slow); 
  void      threadedFreeFermion    (Fermion_t handle);


  void precisionChange (Fermion_t in, Fermion_t out, int control,int cb);

  void impexFermion  (QDPdouble *psi, Fermion_t handle, int import, int cb,int s) ;
  void      importFermion  (QDPdouble *psi, Fermion_t handle,int cb);
  void      exportFermion  (QDPdouble *psi, Fermion_t handle,int cb);
  virtual void importGauge(QDPdouble *gauge, int dir) ;
  
  virtual int doubleStored(void);

  /*Linalg*/
  void      master_fill(Fermion_t psi,double a);
  double    norm(Fermion_t psi);
  void      fill(Fermion_t psi,double a);
  void      set_zero(Fermion_t psi);
#ifdef BFM_QDP_BINDING
  complex<double> inner(Fermion_t x,Fermion_t y);
  double   inner_real(Fermion_t x_t, Fermion_t y_t);
#endif
  void      scale(Fermion_t psi,double a);
  void      scale(Fermion_t psi,double ar,double ai);
  void      conj(Fermion_t psi);
  void      copy(Fermion_t psi,Fermion_t copyme);
  void      axpy(Fermion_t result, Fermion_t x, Fermion_t y,double a);
  void      axpby(Fermion_t result, Fermion_t x, Fermion_t y,double a,double b);
  double    axpy_norm(Fermion_t result, Fermion_t x, Fermion_t y,double a);
  double    cg_psi_p (Fermion_t psi, Fermion_t p, Fermion_t r,Fermion_t mmp,double a,double b);
  double    axpby_norm(Fermion_t result, Fermion_t x, Fermion_t y,double a,double b);
  void      chiralproj_axpby(Fermion_t y, Fermion_t x, double a, int sign, int s_slices);
  void      axpibg5x(Fermion_t r, Fermion_t x, double a,double b);
  void      merge(Float *r, Float *lo, Float *hi, int nsite);
  void four_to_five(Fermion_t input[2], Fermion_t result5d[2]);
  void five_to_four(Fermion_t input5d[2], Fermion_t result[2], int sgn);


  /*Checksumming*/
  uint64_t  checksum(Fermion_t psi);
  uint64_t  checksumGauge(void);
  uint64_t  saved_checksum;

  void dump(Fermion_t psi, char *filename, double scale,int cb);

  ////////////////////////////////////////////////////////////////////////////////////
  // Preconditioned CG My solver of choice (So far)
  ////////////////////////////////////////////////////////////////////////////////////
  int CGNE_M(Fermion_t sol_guess[2], Fermion_t source[2]);
  int CGNE_M(Fermion_t sol_guess[2], Fermion_t source[2], multi1d<bfm_fermion> &evecs, multi1d<double> &evals);
  int CGNE_Mdag(Fermion_t sol_guess[2], Fermion_t source[2]);
  int CGNE_prec_MdagM(Fermion_t sol_guess, Fermion_t source);
  int CGNE_prec_MdagM_multi_shift(Fermion_t sol_guess[], 
				  Fermion_t source,
				  double    shifts[],
				  double    alpha[],
				  int       nshift,
				  double mresidual[],
				  int single); // sum single result

  int CGNE_MdagM(Fermion_t sol_guess[2], Fermion_t source[2]);//Not implemented
  int CGNE(Fermion_t sol_guess[2], Fermion_t source[2]);// deprecated
  int CGNE(Fermion_t sol_guess[2], Fermion_t source[2], multi1d<bfm_fermion> &evecs, multi1d<double> &evals);// deprecated
  int CGNE_prec(Fermion_t sol_guess, Fermion_t source); // Internal


  int CGNE_PV_M(Fermion_t solution[2], Fermion_t source[2],int n, int L5);
  int CGNE_PV_Mdag(Fermion_t solution[2], Fermion_t source[2],int n, int L5);
  int CGNE_PV_prec_M(Fermion_t psi, Fermion_t src,int n, int L5);
  int CGNE_PV_prec(Fermion_t psi, Fermion_t src,int n, int L5);
  /////////////////////////////////////////////
  // Both parity linalg for unprec CG support
  //////////////////////////////////////////////
  double norm(Fermion_t vec,int s);
  double norm(Fermion_t vec[2]){
    return norm(vec[0])+norm(vec[1]);
  }
  double norm(Fermion_t vec[2],int s){
    return norm(vec[0],s)+norm(vec[1],s);
  }
  void axpy(Fermion_t z[2],Fermion_t x[2],Fermion_t y[2],double a){
    axpy(z[0],x[0],y[0],a);
    axpy(z[1],x[1],y[1],a);
  }
  void axpby(Fermion_t z[2],Fermion_t x[2],Fermion_t y[2],double a,double b){
    axpby(z[0],x[0],y[0],a,b);
    axpby(z[1],x[1],y[1],a,b);
  }
  void scale(Fermion_t z[2],double a){
    scale(z[0],a);
    scale(z[1],a);
  }
  int CGNE_Mdag_unprec(Fermion_t sol_guess[2], Fermion_t source[2]);
  int CGNE_MdagM_unprec(Fermion_t sol_guess[2], Fermion_t source[2]);
  int CGNE_M_unprec(Fermion_t sol_guess[2], Fermion_t source[2]);
  int CGNE_unprec(Fermion_t sol_guess[2], Fermion_t source[2]);

  //////////////////////////////////////////
  // Action dependent
  // - ideally would move to have these virtual and
  //   use derived classes for different actions
  //////////////////////////////////////////
  void Munprec(Fermion_t chi[2], 
	       Fermion_t psi[2], 
	       Fermion_t tmp,
	       int dag) ;

  virtual double Mprec(Fermion_t chi, 
		       Fermion_t psi, 
		       Fermion_t tmp, 
		       int dag,int donrm=0) ;

  virtual double MprecTilde(Fermion_t chi, 
		       Fermion_t psi, 
		       Fermion_t tmp, 
		       int dag,int donrm=0) ;

  void Meo(Fermion_t chi, 
	   Fermion_t psi,
	   int result_cb,
	   int dag);

  void Mooee(Fermion_t chi, 
	     Fermion_t psi, 
	     int dag) ;

  void MooeeInv(Fermion_t chi, 
			Fermion_t psi, 
			int dag) ;


  /*Mooee depends on 4d/5d precon versus */
  /*bagel_idx and bagel_idx5d depend on 4d/5d precon versus */
  /*dslash depends (weakly) on 4d/5d precon versus */

  void Mooee5dprec(Fermion_t psi, 
	                   Fermion_t chi, 
	                   int dag);
  void MooeeInv5dprec(Fermion_t psi, 
	            Fermion_t chi, 
	            int dag);
  void Mooee4dprec(Fermion_t psi, 
		   Fermion_t chi, 
		   int dag);
  void MooeeInv4dprec(Fermion_t psi, 
	            Fermion_t chi, 
	            int dag);

  //////////////////////////////////////////
  // PV Trick
  //////////////////////////////////////////
  void MunprecTW(Fermion_t chi[2], 
	       Fermion_t psi[2], 
	       Fermion_t tmp,
	       int n,int L5,
	       int dag) ;

 double MprecTW(Fermion_t chi, 
	      Fermion_t psi, 
	      Fermion_t tmp, int n, int L5,
	      int dag,int donrm=0) ;

  void Mooee5dprec_TW(Fermion_t psi_t, 
				  Fermion_t chi_t, int n, int L5,
				  int dag);

  void MooeeInv5dprec_TW(Fermion_t psi_t, 
				     Fermion_t chi_t, int n, int L5, 
				     int dag);

  //////////////////////////////////////////


  integer chroma_idx(int site[4],int reim,int i, int i_size) ;
  int mprecFlopsPerSite ( void ) ;
  int axpyBytes ( void ) ;
  int axpyNormFlops ( void ) ;
  int axpyFlops ( void ) ;
  int mprecFlops( void ) ;
  int cbSites( void ) ;


/********************************************************************************
 * Public Methods that are specific to certain implementations are virtual
 ********************************************************************************
 */

  virtual void init(bfmarg &arg) = 0; 
  virtual void end (void) = 0;

  virtual integer bagel_idx   (int site[4],int reim,int i, int i_size,int half_cb) = 0;
  virtual integer bagel_idx5d (int site[4],int s,int reim,int i, int i_size,int half_cb) =0 ;
  virtual integer bagel_idx5d (int site,int s,int i_size) =0;

  virtual void dslash(Fermion_t chi, 
		      Fermion_t psi, 
		      int cb,
		      int dag) = 0;

  virtual double dslash_generic(Fermion_t chi, 
		      Fermion_t psi_in, 
		      Fermion_t psi_out, 
		      int cb,
		      int dag,int dperp=1, int addscale=0, double a=0,double b=0) =0 ;

  virtual int simd(void) = 0;
  virtual void pointers_init(void) = 0;
  virtual void pointers_end(void) = 0; 
  virtual void comm_init(void) = 0;
  virtual void comm_end(void) = 0;
  virtual void comm(int result_cb, Fermion_t psi,int dag) = 0; 
  virtual void comm_start(int result_cb, Fermion_t psi,int dag) = 0; 
  virtual void comm_complete(int result_cb, Fermion_t psi) = 0; 
  virtual void comm_gsum(double &val);
  virtual void comm_csum(uint64_t &csum) {}; // Place holder
  virtual bool isBoss();
  virtual int  SPIcomms(void) { return 0; };

};

// Other prototypes
#include <bfm_vmx.h>
void *bfm_alloc(size_t size,int mem_type = mem_slow);
void  bfm_free (void *ptr);

#ifdef BFM_QMP
#include <bfmdop.h>
#include <bfmcommqmp.h>
#ifndef BFM_BGQ_SPI_DSLASH
typedef bfmcommQMP<double> bfm;
typedef bfmcommQMP<double> bfm_dp;
typedef bfmcommQMP<float>  bfm_sp;
template <class Float>
class bfm_internal : public bfmcommQMP<Float> { 
};
#else
#include <bfmcommspi.h> // Override to use SPI comms inside dslash
typedef bfmcommspi<double> bfm;
typedef bfmcommspi<double> bfm_dp;
typedef bfmcommspi<float>  bfm_sp;
template <class Float>
class bfm_internal : public bfmcommspi<Float> { 
};
#endif
#endif

#ifdef BFM_FAKE
#include <bfmdop.h>
#include <bfmcommfake.h>
typedef bfmcommfake<double> bfm;
typedef bfmcommfake<double> bfm_dp;
typedef bfmcommfake<float>  bfm_sp;
template <class Float>
class bfm_internal : public bfmcommfake<Float> { 
};
#endif

#endif
