/*
 * Can I come up with a lock-free implementation that is the
 * same for pthread and BGQ? Need N+1
 */

#include <bfm.h>
#include <stdio.h>


#if 0
#include <firmware/include/Firmware.h>
#include <hwi/include/bqc/A2_core.h>
#include <hwi/include/bqc/A2_inlines.h>
#else
// Warning: Most SPRs are privileged.
// Note: Must be a define because SPRN is encoded in the instruction.
#define mfspr( SPRN )\
({\
   uint64_t tmp;\
   do {\
      asm volatile ("mfspr %0,%1" : "=&r" (tmp) : "i" (SPRN) : "memory" );\
      }\
      while(0);\
   tmp;\
})
#define SPRN_PIR (0x11E)    //Processor ID Register (See GPIR) Hypervisor Read-only
#define SPRN_SPRG7RO (0x107)//Copy of Processor ID Register (See GPIR) Hypervisor Read-only
#define PIR_REG SPRN_SPRG7RO

// read my Thread ID (0 to 3) from SPRN_PIR[ PIR_TID ]
inline int PhysicalThreadID( void )
{
   uint64_t pir = mfspr( PIR_REG );

   return( (int)(pir & 0x3) );
}
#endif

#define BGQ_ATOMIC_BASE_ADDR         (0x04000000000ULL)
#define MY_PHYMAP_MINADDR_CATOMICOPS (0x10000000000ULL)
#define ATOMIC_BASE MY_PHYMAP_MINADDR_CATOMICOPS

#define	ATOMIC_LOAD		(0)
#define	ATOMIC_LOAD_CLEAR	(1<<3)
#define	ATOMIC_LOAD_INC		(2<<3)
#define ATOMIC_LOAD_DEC		(3<<3)
#define	ATOMIC_STORE		(0)
#define	ATOMIC_STORE_ADD	(2)

inline void msync(void) { 
  do { asm volatile ("msync" : : : "memory"); } while(0); 
}

inline uint64_t volatile *Atomic_Alias(uint64_t * volatile addr)
{
  volatile uint64_t* ptr = 
    (uint64_t*) (
		 ATOMIC_BASE
		 |((uint64_t)addr << 5)
		);

  return ptr;
}

inline uint64_t Fetch(uint64_t addr, uint64_t op)
{
  uint64_t tid = PhysicalThreadID() << 6;
  register volatile uint64_t *ptr = (uint64_t *)((addr | op)^tid);
  uint64_t val = *ptr;
  return val;
}

inline void Store(uint64_t addr, uint64_t op, uint64_t value)
{
  uint64_t tid = PhysicalThreadID() << 6;
  register volatile uint64_t *ptr = (uint64_t *)((addr | op)^tid);
  *ptr = value;
}

//uint64_t  ThreadModelFirmware::barrier_step;
//uint64_t  ThreadModelFirmware::barrier_threads;

void ThreadModelFirmware::thread_init_noalloc  (int nthr) 
{
  nthread = nthr;

  atomic_barrier_step    =(uint64_t) Atomic_Alias(&barrier_step);
  atomic_barrier_threads =(uint64_t) Atomic_Alias(&barrier_threads);
}

void ThreadModelFirmware::thread_init  (int nthr)  
{
  thread_init_noalloc  ( nthr) ;
  sum_array = (double *)malloc(nthr*sizeof(double)); // Missing free
}

 
int  ThreadModelFirmware::thread_barrier(void) 
{
  msync();
  
  uint64_t step = Fetch(atomic_barrier_step,ATOMIC_LOAD);
  uint64_t me   = Fetch(atomic_barrier_threads, ATOMIC_LOAD_INC);
  if ( me == nthread-1 ) {
    step++;
    Store(atomic_barrier_threads,ATOMIC_STORE,0x0); 
    Store(atomic_barrier_step,ATOMIC_STORE,step);
  } else { 
    int i=0;
    while( Fetch(atomic_barrier_step,ATOMIC_LOAD) == step ) {
      if ( i == 10000000 ) {
	i=0;
	printf("Thread %d spinning on barrier %d %lx\n",me,step,atomic_barrier_step);
      } else { 
	i++;
      }
    }
  }
  return (int) me;
}

/*This could be shared*/
void ThreadModelFirmware::thread_sum    (double &val,int me) 
{
  sum_array[me] = val;
  val=0;
  thread_barrier();
  for(int i=0;i<nthread;i++) val+= sum_array[i];
  thread_barrier();
}

void ThreadModelSPI::thread_csum    (uint64_t &val,int me) 
{
  uint64_t *csum_array = (uint64_t *)sum_array;
  csum_array[me] = val;
  val=0;
  thread_barrier();
  for(int i=0;i<nthread;i++) val^= csum_array[i];
  thread_barrier();
}

void  *ThreadModelFirmware::thread_bcast(int me,void * val) 
{ 
  if (me == 0) bcast_ptr = val;
  thread_barrier();
  val = bcast_ptr;
  thread_barrier();
  return val; 
};

