/****************************************************************************/
/* PAB Dec 2007                                                             */
/*
  // Methods implemented here
  virtual void init(bfmarg &arg) ; 
  virtual void end (void) ;

  virtual integer bagel_idx   (int site[4],int reim,int i, int i_size,int half_cb) = 0;
  virtual integer bagel_idx5d (int site[4],int s,int reim,int i, int i_size,int half_cb) =0 ;
  virtual integer bagel_idx5d (int site,int s,int i_size) =0;

  virtual void dslash(Fermion_t chi, 
		      Fermion_t psi, 
		      int cb,
		      int dag,int dperp=1) ;


  virtual void Dperp(Fermion_t psi, 
		     Fermion_t chi, 
		     int cb,
		     int dag);

  virtual void pointers_init(void) ;
  virtual void pointers_end(void) ; 
*/
/****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h> 
#include <fcntl.h>


#include "bfm.h"
#include <sys/time.h>

static double complex_i_simd1[32] __attribute__ ((aligned(32)))  = { 0.0, 1.0, 
                                                                     0.0,-1.0, 
                                                                     0.0, 0.0, 
                                                                     2.0, 2.0, 
								     0.0, 0.0,
								     0.0, 0.0,
								     0.0, 0.0,
								     0.0, 0.0
};

static double complex_i_simd2[32] __attribute__ ((aligned(32))) = { 0.0, 1.0, 0.0, 1.0, //0-3
                                                                    0.0,-1.0, 0.0,-1.0, //4-7
                                                                    0.0, 0.0, 0.0, 0.0, //8-11 Zero
								    2.0, 2.0, 2.0, 2.0, //12-15
                                                                    0.0, 0.0, 0.0, 0.0, //16-19 Neg2mass
                                                                    0.0, 0.0, 0.0, 0.0, //20-23 A
                                                                    0.0, 0.0, 0.0, 0.0, //24-27 B
                                                                    0.0, 0.0, 0.0, 0.0  //28-32 Dot
};

template <class Float>
int bfmdop<Float>::simd(void)
{
 return BAGEL_NSIMD;
}

int shm_fd;

template <class Float>
void bfmdop<Float>::init(bfmarg &arg)  
{
  int words;             /* size of the spinor field on the         */
  	  	  	 /* sublattice checkerboard                 */

  int slx;                      /* x-direction size of node sublattice */
  int sly;                      /* y-direction size of node sublattice */
  int slz;                      /* z-direction size of node sublattice */
  int slt;                      /* t-direction size of node sublattice */
  int i;
  int mu;

/*--------------------------------------------------------------------------*/
/* Set sublattice direction sizes                                           */
/*--------------------------------------------------------------------------*/
  * ((bfmarg *)this) = arg;


  /*-----------------------------------------------------------------
   * Advise how many threads will call numerical operations
   *-----------------------------------------------------------------
   */
  if ( this->isBoss() ) 
    printf("bfmdop:: Initialising BAGEL-2 solver %d %d for %d threads\n",
	   this->verbose,
	   this->isBoss(),this->threads); fflush(stdout);

  thread_init(this->threads);  

  
  this->halo_depth=1;
  if ( this->solver == WilsonNN ) { 
    this->halo_depth =2;
  }

  /*-------------------------------------------------------------
   * Introduce 5d precon ; use ncb to refer to 4d checkerboards
   *-------------------------------------------------------------
   */
  if ( this->precon_5d && (this->Ls==1) ) { 
    if ( this->isBoss() ) printf("bfm::init: precon_5d & Ls=1 logic bomb\n");
    exit(-1);
  }
  if ( this->precon_5d ) this->ncb=1;
  else this->ncb=2;
  if ( this->precon_5d ) this->cbLs=this->Ls/2;
  else this->cbLs=this->Ls;

  this->node_cbvol = this->node_latt[0]
                               *this->node_latt[1]
                               *this->node_latt[2]
                               *this->node_latt[3]
                               /this->ncb;

  /***************************************************/
  /*This should be selected depending on architecture*/
  if ( this->verbose && this->isBoss() ) printf("bfm::init: nsimd =%d BAGEL2_NSIMD=%d\n",simd(),BAGEL_NSIMD);
  this->nsimd = this->simd();
  this->simd_grid[0] = 1; // Query. Is it possible to just move this around?
  this->simd_grid[1] = 1;
  this->simd_grid[2] = 1;
  this->simd_grid[3] = this->nsimd;
  if (this->nsimd == 1 )      complex_i_simd = complex_i_simd1;
  else if (this->nsimd == 2 ) complex_i_simd = complex_i_simd2;
  else exit(0);
  /***************************************************/

  this->simd_latt[0] = slx = this->node_latt[0]/this->simd_grid[0];
  this->simd_latt[1] = sly = this->node_latt[1]/this->simd_grid[1];
  this->simd_latt[2] = slz = this->node_latt[2]/this->simd_grid[2];
  this->simd_latt[3] = slt = this->node_latt[3]/this->simd_grid[3];

  this->base_parity =(this->ncoor[0]*this->node_latt[0] 
              + this->ncoor[1]*this->node_latt[1]
              + this->ncoor[2]*this->node_latt[2]
              + this->ncoor[3]*this->node_latt[3])&0x1;

/*--------------------------------------------------------------------------*/
/* Set periodic wrap back or not                                            */
/*--------------------------------------------------------------------------*/
  this->local_comm[0] = arg.local_comm[0];
  this->local_comm[1] = arg.local_comm[1];
  this->local_comm[2] = arg.local_comm[2];
  this->local_comm[3] = arg.local_comm[3];

/*-----------------------------------------------------------------------*/
/* compute the subgrd volume of each chkbd ... at least two local dims   */
/* must be even for this code to be correct.                             */
/*-----------------------------------------------------------------------*/
  this->simd_cbvol = (slx * sly * slz * slt)/this->ncb;
  
  this->simd_nbound[0] = this->halo_depth*(sly * slz * slt)/this->ncb; 
  this->simd_nbound[1] = this->halo_depth*(slx * slz * slt)/this->ncb;
  this->simd_nbound[2] = this->halo_depth*(slx * sly * slt)/this->ncb;
  this->simd_nbound[3] = this->halo_depth*(slx * sly * slz)/this->ncb;

  this->simd_allbound  = this->simd_nbound[0]
                       + this->simd_nbound[1]
                       + this->simd_nbound[2]
                       + this->simd_nbound[3];

  if ( this->simd_nbound[0] * slx * this->ncb != (slx*sly*slz*slt)*this->halo_depth ) {
    if ( this->isBoss() ) printf("bfm::init Even x logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[1] * sly * this->ncb != (slx*sly*slz*slt)*this->halo_depth ) {
    if ( this->isBoss() ) printf("bfm::init Even y logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[2] * slz * this->ncb != (slx*sly*slz*slt)*this->halo_depth ) {
    if ( this->isBoss() ) printf("bfm::init Even z logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[3] * slt * this->ncb != (slx*sly*slz*slt)*this->halo_depth ) {
    if ( this->isBoss() ) printf("bfm::init Even t logic bomb\n");
    exit(-1);
  }

  /*------------------------------------------------------------------------*/
  /* Check shape                                                            */
  /*------------------------------------------------------------------------*/
  if ( (slx&1) || (slz&0x1) || (slt&0x1) ) {
    if ( this->isBoss() ) printf("bfm<Float>::init Bagel is refusing to run as a sub latt is odd\n");
    exit(-1);
  }


/*--------------------------------------------------------------------------*/
/* Reserve memory for 1 spatial site of two spinor per thread               */
/*--------------------------------------------------------------------------*/
  words = this->cbLs * HALF_SPINOR_SIZE * this->nsimd * 10 * this->nthread;

  this->two_spinor = (Float *)bfm_alloc(words*sizeof(Float));
  if(this->two_spinor == 0){
    if ( this->isBoss() ) printf("bfm<Float>::init two_spinor allocate\n");
    exit(-1);
  }
  bzero((void*)this->two_spinor, words*sizeof(Float));

/*--------------------------------------------------------------------------*/
/* Reserve memory for gauge field                                           */
/*--------------------------------------------------------------------------*/
  words = GAUGE_SIZE * this->nsimd * this->simd_cbvol *2*2;

  if ( this->SPIcomms() ) {
    if ( this->isBoss() ) {
      printf("bfm::u allocate from shared mem\n"); 
      fflush(stdout);
    }
    shm_fd = shm_open("/bagel-gauge-field",O_RDWR|O_CREAT,0777);
    this->u= (Float *)mmap(NULL,words*sizeof(Float),PROT_READ|PROT_WRITE,MAP_PRIVATE,shm_fd,0);
    if(this->u == (Float *)MAP_FAILED){
      if ( this->isBoss() ) printf("bfm::u allocate\n");
      exit(-1);
    }
  } else { 
    if ( this->isBoss() ) {
      printf("bfm::u allocate from heap\n"); 
      fflush(stdout);
    }
    this->u = (Float *)bfm_alloc(words*sizeof(Float));
    if(this->u == 0){
      if ( this->isBoss() ) printf("bfm::u allocate\n");
      exit(-1);
    }
  }

/*----------------------------------------------------------------------*/
/* Build the pointer table                                              */
/*----------------------------------------------------------------------*/
  this->pointers_init();
  
/*----------------------------------------------------------------------*/
/* Initialise the comms                                                 */
/*----------------------------------------------------------------------*/
  this->comm_init();

  // Possible Overlap implementation init
  this->GeneralisedFiveDimInit();
}

template <class Float>
void bfmdop<Float>::end (void)
{
  this->comm_end();
  this->pointers_end();
  bfm_free(this->two_spinor);
  if ( this->SPIcomms() ) {
    uint64_t words = GAUGE_SIZE * this->nsimd * this->simd_cbvol *2*2;
    munmap(this->u,words*sizeof(Float));
    close(shm_fd);
  } else { 
    bfm_free(this->u);
  }
}

template <class Float>
integer bfmdop<Float>::bagel_idx5d (int site,int s,int nspinco) 
{
  if ( this->precon_5d ) s = s/2;
  return s*nspinco*this->simd()*2 + site*nspinco*this->simd()*2*this->cbLs;
}
template <class Float>
integer bfmdop<Float>::bagel_idx5d (int x[4],int s,int reim,int i, int i_size, int half_cb)
{
  if ( this->precon_5d ) s = s/2;
  return this->bagel_idx(x,reim,i+s*i_size,i_size*this->cbLs,half_cb);
}
template <class Float>
integer bfmdop<Float>::bagel_idx (int x[4],int reim,int i, int i_size, int half_cb)
{
  int gx[4];
  int mu;

  for (mu=0;mu<4;mu++ ) gx[mu] = x[mu]/this->simd_latt[mu];

  integer simd_el = gx[0]+this->simd_grid[0]*( gx[1]
                         +this->simd_grid[1]*( gx[2]
                         +this->simd_grid[2]*  gx[3] ));
  int scoor[4]; // SIMD coordinate

  scoor[0] = x[0] % this->simd_latt[0]; /*Work out our location within the SIMD vol*/
  scoor[1] = x[1] % this->simd_latt[1];
  scoor[2] = x[2] % this->simd_latt[2];
  scoor[3] = x[3] % this->simd_latt[3];

  int scb   = this->parity(scoor);          /*Work out local checkerboard of site*/
  int ssite = this->psite(scoor,this->simd_latt); /*Work out checkerboarded site number*/

  // Real/Imaginary innermost
  int idx = reim + simd_el*2
                 + i*simd()*2
      + ssite*i_size*simd()*2;

// Bagel only deals with half checkerboard fermion vectors
// to allow for red black solution
  if ( !half_cb ) idx+= scb*this->simd()*2*i_size*this->simd_cbvol;

  return idx;
}

/*
 * Implement all the platform specific routines for BG/L,P,Q
 * These are:
 *  bagel_idx -- for mapping the layout in import/export routines
 *  dslash    -- calls the BG assembler
 *  Mooee     -- preconditioning is aware of the data layout 
 *  MooeeInv  -- preconditioning is aware of the data layout 
 *  comm_init,comm_end,comm
 */

static double two_mass_simd1[4] __attribute__ ((aligned(32)));
static double two_mass_simd2[8] __attribute__ ((aligned(32)));


/*
 * Temporarily revert experiment while 
 * I figure out twisted
template <class Float>
double bfmdop<Float>::Mprec(Fermion_t psi, 
			       Fermion_t chi,
			       Fermion_t tmp, 
			       int dag,int do_nrm)
{
    int dperp_yes=1;
    int axpy_yes=1;
    int axpy_no=0;

    double nrm;
    double     ident = (5.-this->M5);
    double       hop = -0.25/(ident);
    double ident_hop = ident/hop;
    // tmp= d psi
    this->dslash_generic(psi,tmp,tmp,Even,dag,dperp_yes,axpy_no ,0.0);
    // chi= d d psi + ident * psi
    nrm = this->dslash_generic(tmp,psi,chi,Odd ,dag,dperp_yes,axpy_yes,ident_hop,hop);
    return nrm;
}
*/

template <class Float>
void bfmdop<Float>::dslash(Fermion_t psi, // Input
			      Fermion_t chi, // Result
			      int cb,
			      int dag)

{
  double z=0.0;
  double one=1.0;
  int dperp;
  if ( this->solver == DWF ) dperp = 1;
  else dperp=0;
  dslash_generic(psi,chi,chi,cb,dag,dperp,0,z,one);
}

#define OVERLAP_COMMS_COMPUTE
#ifdef OVERLAP_COMMS_COMPUTE
 
template <class Float>
double bfmdop<Float>::dslash_generic(Fermion_t psi, // Input
              	    Fermion_t chi_in,  // Axpy
              	    Fermion_t chi_out, // Result
                    int cb,
                    int dag,
		    int dperp,
		    // if true, chi =b ( a chi + dslash psi )
    	            int addscale ,
  	            double a,
		    double b)
{

  /*-----------------------------------------------------
   *Divide up work between threads
   *-----------------------------------------------------
   */
  int me, thrvol,throff;
  
  thread_work(this->simd_cbvol,me,thrvol,throff);

  /*Turn global checkerboard into the checkerboard within this node*/
  int lcb = (cb + this->base_parity) & 1;
  cb = lcb;

  uint64_t t1,t2;
  /*Comms & D5 */
  t1 = GetTimeBase();
  this->comm_start(lcb,psi,dag);
  t2 = GetTimeBase();
  uint64_t t_comm_start=t2-t1;

  //////////////////////////////////////////////////////////
  // Have to put these on stack for dot product even though it is ugly
  // CPS style DWF implementation
  // possible B(AX+Y) operation
  /////////////////////////////////////////////////////////
  double stack_i_simd[32] __attribute__ ((aligned(32)));

  if      (this->nsimd == 1 ) {
    for(int n=0;n<32;n++) stack_i_simd[n] = complex_i_simd1[n];
    stack_i_simd[8] = stack_i_simd[9] = -2*this->mass;
    stack_i_simd[10]= stack_i_simd[11]  = a;
    stack_i_simd[12]= stack_i_simd[13]  = b;
  } else if (this->nsimd == 2 ) {
    for(int n=0;n<32;n++) stack_i_simd[n] = complex_i_simd2[n];
    stack_i_simd[16]= stack_i_simd[17]= -2*this->mass;
    stack_i_simd[18]= stack_i_simd[19]= -2*this->mass;
    stack_i_simd[20]= stack_i_simd[21]  = a;
    stack_i_simd[22]= stack_i_simd[23]  = a;
    stack_i_simd[24]= stack_i_simd[25]  = b;
    stack_i_simd[26]= stack_i_simd[27]  = b;
  }
  
  /*
   * Setup gauge fields for 5d/4d DWF or Wilson
   */
  Float *gauge_par;
  Float *utmp = (Float *)this->u;
  if ( cb == 1 ) { 
    gauge_par = utmp+ (2*GAUGE_SIZE*this->simd_cbvol*this->simd()); 
  } else { 
    gauge_par = utmp;
  }
  if ( this->precon_5d ) { 
    gauge_par = utmp;
  } 

  bgq_l1p_optimisation(1);

  integer args[10];
  integer arg_p = (integer)args; 
  args[0] = ((integer)chi_in) + throff
                         * this->cbLs
                         * this->nsimd
                         * 24*sizeof(Float);
  args[1] = (integer)psi;
  args[2] = (integer)(gauge_par + throff*18*8*this->simd());
  args[3] = (integer)thrvol;
  args[4] = (integer)this->cbLs;
  args[5] = (integer) & this->shift_table[cb][throff*16];
  args[6] = (integer)stack_i_simd;
  args[7] = ((integer)this->two_spinor) + me 
                                        * sizeof(Float)
                                        * this->cbLs 
                                        * HALF_SPINOR_SIZE * this->nsimd * 10 ;
  if ( this->SPIcomms() ) 
    args[8] = (integer)recvbufs[0];
  else 
    args[8] = (integer)psi;

  args[9] = ((integer)chi_out) + throff
                         * this->cbLs
                         * this->nsimd
                         * 24*sizeof(Float);

  int single =  1;
  if ( sizeof(Float) == sizeof(double) ) single = 0;

  uint64_t t_comm_complete=0;
  uint64_t t_Dslashext    =0;
  uint64_t t_Dslash       =0;

  this->thread_barrier();
  t1 = GetTimeBase();
  if ( dperp ) {
    if ( addscale ) {
      if ( dag ) {
        if (single)   vmx_dwf_dp_a_int_dag_s(arg_p);
        else          vmx_dwf_dp_a_int_dag(arg_p);
      } else {
        if ( single ) vmx_dwf_dp_a_int_s(arg_p);
        else          vmx_dwf_dp_a_int(arg_p);
      }
    } else { 
      if ( dag ) {
	if (single)   vmx_dwf_dp_int_dag_s(arg_p);
	else          vmx_dwf_dp_int_dag(arg_p);
      } else {
	if ( single ) vmx_dwf_dp_int_s(arg_p);
	else          vmx_dwf_dp_int(arg_p);
      }
    }
  } else {
    if ( addscale ) {
      if ( dag ) {
        if (single)   vmx_wil_a_int_dag_s(arg_p);
        else          vmx_wil_a_int_dag(arg_p);
      } else {
        if ( single ) vmx_wil_a_int_s(arg_p);
        else          vmx_wil_a_int(arg_p);
      }
    } else { 
      if ( dag ) {
	if (single)   vmx_wil_int_dag_s(arg_p);
	else          vmx_wil_int_dag(arg_p);
      } else {
	if ( single ) vmx_wil_int_s(arg_p);
	else          vmx_wil_int(arg_p);
      }
    }
  }
  this->thread_barrier();
  t2 = GetTimeBase();
  t_Dslash = t2-t1;

  t1 = GetTimeBase();
  this->comm_complete(lcb,psi);
  t2 = GetTimeBase();
  t_comm_complete=t2-t1;
  
  args[0] = args[9];
  t1 = GetTimeBase();
  if ( addscale ) {
    if ( dag ) {
      if (single)   vmx_dwf_ext_scale_dag_s(arg_p);
      else          vmx_dwf_ext_scale_dag(arg_p);
    } else {
      if ( single ) vmx_dwf_ext_scale_s(arg_p);
      else          vmx_dwf_ext_scale(arg_p);
    }
  } else { 
    if ( dag ) {
      if (single)   vmx_dwf_ext_dag_s(arg_p);
      else          vmx_dwf_ext_dag(arg_p);
    } else {
      if ( single ) vmx_dwf_ext_s(arg_p);
      else          vmx_dwf_ext(arg_p);
    }
  }
  t2 = GetTimeBase();
  t_Dslashext = t2-t1;

  double dot;
  if      (this->nsimd == 1 )  {
    dot = stack_i_simd[14]+stack_i_simd[15];
  }  else if (this->nsimd == 2 ) {
    dot = stack_i_simd[28]+stack_i_simd[29]+stack_i_simd[30]+stack_i_simd[31];
  }
  this->thread_sum(dot,me);
  this->comm_gsum(dot);

  // This is painful... all the performance analysis doubles the code length
   if ( this->iter == this->time_report_iter ) {
     
     if ( (me==0) ) {
       if ( this-> isBoss() ) { 
	 
	 int allbound = 0;
	 int maxbound = 0;
	 
	 for (int mu=0; mu<4;mu++)
	   if ( !this->local_comm[mu] ) allbound += this->simd_nbound[mu];
	 for (int mu=0; mu<4;mu++) 
	   if ( (!this->local_comm[mu]) && this->simd_nbound[mu]>maxbound ) 
	     maxbound = this->simd_nbound[mu];
	 
	 double XBytes = 2*allbound * this->simd() * 12* sizeof(Float)*this->cbLs;
	 int simd_gm  = this->simd_cbvol*8;
	 int simd_egm = 2*allbound;
	 int simd_igm = simd_gm-simd_egm;

	 printf("simd_gauge_muls     = %d\n",simd_gm);
	 printf("simd_int_gauge_muls = %d\n",simd_igm);
	 printf("simd_ext_gauge_muls = %d\n",simd_egm);
	 double int_flops = (12+132)*simd_igm+7*24*this->simd_cbvol;
	 double ext_flops = (12+132)*simd_egm;
	 double flops = 1344.*this->simd_cbvol*this->simd()*this->cbLs;
	 
	 printf("int_flops     = %le\n",int_flops);
	 printf("ext_flops     = %le\n",ext_flops);
	 printf("flops         = %le\n",flops);

	 printf("t_comm_start    %lld\n", t_comm_start);
	 printf("t_comm_complete %lld\n", t_comm_complete);
	 printf("comm bytes %le completed in %le usecs\n", XBytes*2.0 ,1600./(t_Dslash+t_comm_start+t_comm_complete+t_Dslashext));

	 printf("t_Dslash        %lld\n", t_Dslash);
	 printf("t_Dslashext     %lld\n", t_Dslashext);
	
	 printf("Pass-1+2   Mflop/s %16.3le\n", flops*1600./(t_Dslash+t_Dslashext));
	 printf("Pass-1     Mflop/s %16.3le\n", int_flops*1600./(t_Dslash));
	 printf("Pass-2     Mflop/s %16.3le\n", ext_flops*1600./(t_Dslashext));

	 printf("Aggregate        Mflop/s %16.3le\n", flops*1600./(t_Dslash+t_comm_start+t_comm_complete+t_Dslashext));

	 fflush(stdout);
       }
     }
     this->thread_barrier();
   }

  bgq_l1p_optimisation(0);

  return dot;
}

#else
 
template <class Float>
double bfmdop<Float>::dslash_generic(Fermion_t psi, // Input
              	    Fermion_t chi_in,  // Axpy
              	    Fermion_t chi_out, // Result
                    int cb,
                    int dag,
		    int dperp,
		    // if true, chi =b ( a chi + dslash psi )
    	            int addscale ,
  	            double a,
		    double b)
{

  /*-----------------------------------------------------
   *Divide up work between threads
   *-----------------------------------------------------
   */
  int me, thrvol,throff;
  
  thread_work(this->simd_cbvol,me,thrvol,throff);

  /*Turn global checkerboard into the checkerboard within this node*/
  int lcb = (cb + this->base_parity) & 1;
  cb = lcb;

  /*Comms & D5 */
  this->comm_start(lcb,psi,dag);
  this->comm_complete(lcb,psi);


  //////////////////////////////////////////////////////////
  // Have to put these on stack for dot product even though it is ugly
  //
  // CPS style DWF implementation
  // possible B(AX+Y) operation
  /////////////////////////////////////////////////////////
  double stack_i_simd[32] __attribute__ ((aligned(32)));

  if      (this->nsimd == 1 ) {
    for(int n=0;n<32;n++) stack_i_simd[n] = complex_i_simd1[n];
    stack_i_simd[8] = stack_i_simd[9] = -2*this->mass;
    stack_i_simd[10]= stack_i_simd[11]  = a;
    stack_i_simd[12]= stack_i_simd[13]  = b;
  } else if (this->nsimd == 2 ) {
    for(int n=0;n<32;n++) stack_i_simd[n] = complex_i_simd2[n];
    stack_i_simd[16]= stack_i_simd[17]= -2*this->mass;
    stack_i_simd[18]= stack_i_simd[19]= -2*this->mass;
    stack_i_simd[20]= stack_i_simd[21]  = a;
    stack_i_simd[22]= stack_i_simd[23]  = a;
    stack_i_simd[24]= stack_i_simd[25]  = b;
    stack_i_simd[26]= stack_i_simd[27]  = b;
  }
  
  /*
   * Setup gauge fields for 5d/4d DWF or Wilson
   */
  Float *gauge_par;
  Float *utmp = (Float *)this->u;
  if ( cb == 1 ) { 
    gauge_par = utmp+ (2*GAUGE_SIZE*this->simd_cbvol*this->simd()); 
  } else { 
    gauge_par = utmp;
  }
  if ( this->precon_5d ) { 
    gauge_par = utmp;
  } 

  bgq_l1p_optimisation(1);

  integer args[10];
  integer arg_p = (integer)args; 
  args[0] = ((integer)chi_in) + throff
                         * this->cbLs
                         * this->nsimd
                         * 24*sizeof(Float);
  args[1] = (integer)psi;
  args[2] = (integer)(gauge_par + throff*18*8*this->simd());
  args[3] = (integer)thrvol;
  args[4] = (integer)this->cbLs;
  args[5] = (integer) & this->shift_table[cb][throff*16];
  args[6] = (integer)stack_i_simd;
  args[7] = ((integer)this->two_spinor) + me 
                                        * sizeof(Float)
                                        * this->cbLs 
                                        * HALF_SPINOR_SIZE * this->nsimd * 10 ;
  if ( this->SPIcomms() ) 
    args[8] = recvbufs[0];
  else 
    args[8] = psi;

  args[9] = ((integer)chi_out) + throff
                         * this->cbLs
                         * this->nsimd
                         * 24*sizeof(Float);

  int single =  1;
  if ( sizeof(Float) == sizeof(double) ) single = 0;

  this->thread_barrier();

  if ( dperp ) {
    if ( addscale ) {
      if ( dag ) {
        if (single)   vmx_dwf_dp_as_dag_s(arg_p);
        else          vmx_dwf_dp_as_dag(arg_p);
      } else {
        if ( single ) vmx_dwf_dp_as_s(arg_p);
        else          vmx_dwf_dp_as(arg_p);
      }
    } else { 
      if ( dag ) {
	if (single)   vmx_dwf_dp_dag_s(arg_p);
	else          vmx_dwf_dp_dag(arg_p);
      } else {
	if ( single ) vmx_dwf_dp_s(arg_p);
	else          vmx_dwf_dp(arg_p);
      }
    }
  } else {
    if ( addscale ) {
      if ( dag ) {
        if (single)   vmx_wil_as_dag_s(arg_p);
        else          vmx_wil_as_dag(arg_p);
      } else {
        if ( single ) vmx_wil_as_s(arg_p);
        else          vmx_wil_as(arg_p);
      }
    } else { 
      if ( dag ) {
	if (single)   vmx_wil_dag_s(arg_p);
	else          vmx_wil_dag(arg_p);
      } else {
	if ( single ) vmx_wil_s(arg_p);
	else          vmx_wil(arg_p);
      }
    }
  }
  this->thread_barrier();
   
  double dot;
  if      (this->nsimd == 1 )  {
    dot = stack_i_simd[14]+stack_i_simd[15];
  }  else if (this->nsimd == 2 ) {
    dot = stack_i_simd[28]+stack_i_simd[29]+stack_i_simd[30]+stack_i_simd[31];
  }
  this->thread_sum(dot,me);
  this->comm_gsum(dot);

  bgq_l1p_optimisation(0);

  return dot;
}
#endif



#if 0
// This routine works out Dperp in DWF
template <class Float>
void bfmdop<Float>::Dperp(Fermion_t psi, 
		   Fermion_t chi, 
		   int cb,
		   int dag)
{
  printf("Deprecated\n");
  exit(-1);
  //Temporary debug hack 
  two_mass_simd1[0] = two_mass_simd1[1] = 2.0;
  two_mass_simd1[2] = two_mass_simd1[3] =-2.0*this->mass ;

  two_mass_simd2[0] = two_mass_simd2[1] = 2.0;
  two_mass_simd2[2] = two_mass_simd2[3] = 2.0;
  two_mass_simd2[4] = two_mass_simd2[5] =-2.0*this->mass ;
  two_mass_simd2[6] = two_mass_simd2[7] =-2.0*this->mass ;

  int me,thrvol,throff;
  thread_work(this->simd_cbvol,me,thrvol,throff);

  integer args[10];
  integer arg_p = (integer)args; 
  int fermoff =  24*throff
                   * this->nsimd
                   * this->cbLs
                   * sizeof(Float);
  args[0] = (integer) psi + fermoff;
  args[1] = (integer) chi + fermoff;
  args[2] = (integer) thrvol;
  args[3] = (integer) this->cbLs;
  args[4] = (integer) this->cb_table[cb] + throff*sizeof(integer);
  args[5] = (integer) complex_i_simd;

  if(this->simd() == 1 ) args[6] = (integer) two_mass_simd1;
  if(this->simd() == 2 ) args[6] = (integer) two_mass_simd2;

  int single = 1;
  if (sizeof(Float)==sizeof(double) ) single = 0; 

  if ( dag ) {
    if (single ) vmx_dperp_dag_s(arg_p);
    else         vmx_dperp_dag(arg_p);
  } else { 
    if (single ) vmx_dperp_s(arg_p);
    else         vmx_dperp(arg_p);
  }

}
#endif


template <class Float>
void bfmdop<Float>::pointers_end(void)
{
  for ( int cb = 0;cb<Ncb;cb++ ) {
    bfm_free(this->shift_table[cb]);
    bfm_free(this->cb_table[cb]);
  }
  for(int cb=0;cb<this->ncb;cb++) {
    for(int pm=0;pm<2;pm++) {
      for(int mu=0; mu<4;mu++) {
	bfm_free(this->face_table[cb][pm][mu]);
      }
    }
  }
}


/*Layout is:                                     */
/* PSI   [site/Nsimd][Ls][spin][co][reim][Nsimd] */
/* Gauge [site/Nsimd][mu][co][co][reim][Nsimd]   */
/* PSI contains attached face buffers at the back of array */

template <class Float>
void bfmdop<Float>::pointers_init(void)
{

  int mu;
  int cb, pm;
  int shift;
  int x,y,z,t;
  int lx,ly,lz,lt;
 
  int bound_index[Ncb][NMinusPlus][ND];
  int local_p;
  int local_addr[ND];
  int shift_addr[ND];

  int cbsite,shift_cbsite;
  int offset,tab,table_size;

  lx = this->simd_latt[0];
  ly = this->simd_latt[1];
  lz = this->simd_latt[2];
  lt = this->simd_latt[3];

  /////////////////////////////////
  //Next nearest wilson checks
  /////////////////////////////////
  if ( this->solver == WilsonNN ) {
    if ( this ->halo_depth != 2 ) {
      printf("Halo depth !=2 and WilsonNN\n"); 
      exit(-1);
    }
  } else if ( this ->halo_depth != 1 ) {
    printf("Halo depth >1 and not WilsonNN");
  }
  //////////////////////

  for ( cb = 0 ; cb<Ncb ; cb++) {
    table_size = NMinusPlus*(this->simd_cbvol+1)*2*Nmu;
    this->shift_table[cb]=(short_integer *)bfm_alloc( table_size*sizeof(short_integer));
    for(int i=0;i<table_size;i++) this->shift_table[cb][i]=0;
  }
  for ( cb = 0 ; cb<Ncb ; cb++) {
    table_size = (this->simd_cbvol+8);
    this->cb_table[cb]=(integer *)bfm_alloc( table_size*sizeof(integer));
  }

  for(cb=0;cb<this->ncb;cb++){
    for(pm=0;pm<2;pm++){
      for(mu=0; mu<4;mu++){

	this->face_table[cb][pm][mu]= (integer *)
	     bfm_alloc(this->simd_allbound*sizeof(integer)*this->halo_depth);
	if ( this->ncb==1 ) this->face_table[1][pm][mu]=NULL;

      }
    }
  }
  
  for( mu = 0 ; mu < Nmu ; mu++ ) {
    bound_index[0][0][mu] = 0;
    bound_index[0][1][mu] = 0;
    bound_index[1][0][mu] = 0;
    bound_index[1][1][mu] = 0;
  }

  /*
   * This bit points to the first element after the end of the
   * body for comm_offset[Minus][0] 
   */

  if ( this->SPIcomms() ) {
    if ( this->isBoss() ) printf("bfmdop:: Configuring for SPI comms\n");
    this->comm_offset[Minus][0]=0;
  } else {
    if ( this->isBoss() ) printf("bfmdop:: Configuring for QMP comms\n");
    this->comm_offset[Minus][0]  = this->simd_cbvol; 
  }

  /* Now we do the rest of the comms offsets packing densely.
     nbound[mu-1] sites after the previous */
  this->comm_offset[Minus][1]  = this->comm_offset[Minus][0] + this->simd_nbound[0] ;
  this->comm_offset[Minus][2]  = this->comm_offset[Minus][1] + this->simd_nbound[1] ;
  this->comm_offset[Minus][3]  = this->comm_offset[Minus][2] + this->simd_nbound[2] ;

  /* Now the PLUS directions. They are the same as the minus directions 
     pushed by allbound */
  this->comm_offset[Plus][ 0 ] = this->comm_offset[Minus][0] + this->simd_allbound ;
  this->comm_offset[Plus][ 1 ] = this->comm_offset[Minus][1] + this->simd_allbound ;
  this->comm_offset[Plus][ 2 ] = this->comm_offset[Minus][2] + this->simd_allbound ;
  this->comm_offset[Plus][ 3 ] = this->comm_offset[Minus][3] + this->simd_allbound ;


  /* Now work out the shifts */
  for ( pm = 0; pm<2; pm++ ) {

    /* Shift direction */
    if ( pm == Plus ) { /*forwards or backwards*/
      shift = 1;
    } else { 
      shift = -1;
    }
    shift = shift * this->halo_depth; // Two site hopping

      /*Loop naively over local lattice*/
    for ( t=0 ; t<lt ; t++ ) {
     local_addr[3] = t;
     for ( z=0 ; z<lz ; z++ ) {
      local_addr[2] = z;
      for ( y=0 ; y<ly ; y++ ) {
       local_addr[1] = y;
       for ( x=0 ; x<lx ; x++ ) {
        local_addr[0] = x;

	/* local_addr contains the coordinates of the point */
	/* we get the parity of the point                   */
        local_p = this->parity(local_addr);
	/* get the cbsite of the point for within parity */
        cbsite   = this->psite(local_addr,this->simd_latt);


	// Used for DWF Dperp
	this->cb_table[0][cbsite] = (x+y+z+t)&0x1;
	this->cb_table[1][cbsite] = 1-this->cb_table[0][cbsite];

	// Store info for the dslash as want to integrate the 5d hopping term
	int info_idx;
	info_idx = this->interleave_site(1,5,cbsite);
	this->shift_table[0][info_idx] =   this->cb_table[0][cbsite];
	this->shift_table[1][info_idx] = 1-this->cb_table[0][cbsite];

	info_idx = this->interleave_site(1,4,cbsite);
	this->shift_table[local_p][info_idx] = cbsite*24*this->nsimd*sizeof(Float)*this->cbLs;

        for ( mu = 0 ; mu < 4; mu ++) {

	  /* Get the coordinates of the point shifted in direction 
	     +/- mu. (+/- is encoded in shift, depending on where 
	     we are in the +/- loop) */

          shift_addr[0] = x;
          shift_addr[1] = y;
          shift_addr[2] = z;
          shift_addr[3] = t;

          shift_addr[mu] += shift;

	  /* Get the interleaved site address of the local point
	     (to which we are shifting data for dslash stencil*/
	  tab = this->interleave_site(pm,mu,cbsite);

          /*we're on the boundary*/
	  int on_boundary = 0;
	  if ( (shift_addr[mu] < 0) || (shift_addr[mu]>=this->simd_latt[mu] ) ) {
	    on_boundary	= 1;
          }

          /*
	   * Get the offset to the neighbour site in the spinor with periodicity
	   */
          shift_addr[mu]= (shift_addr[mu] + this->simd_latt[mu])%this->simd_latt[mu];
          shift_cbsite   = psite(shift_addr,this->simd_latt);

	  info_idx = this->interleave_site(0,5,cbsite);
          if ( on_boundary == 0 ) { 

	    //	    this->shift_table[local_p][tab] = shift_cbsite*2;
	    this->shift_table[local_p][tab] = shift_cbsite*24*this->nsimd*sizeof(Float)*this->cbLs;

	  } else {

	    /*
	     * The face_table contains a pointer to the
	     * neighbour site, to be gathered into the comms buffers.
             * local_p is the RESULT checkerboard/parity
	     * This should be sent in the OPPOSITE (1-pm) direction.
	     * Corresponding received data (same if loopback) 
	     * is stored the comm_offset[pm]
	     */
	    int bdx = bound_index[local_p][pm][mu];
	    this->face_table[local_p][pm][mu][bdx] = shift_cbsite;

	    /*
	     * And we use face_table[local_p][pm][mu] in conjunction
	     * with the send buffer[pm][mu] and source parity local_p
	     */
            if ( (!this->local_comm[mu]) || (this->simd_grid[mu]>1) ) {
	      this->shift_table[local_p][tab] = (this->comm_offset[pm][mu]*24 + bdx*12)
		                              * this->nsimd*sizeof(Float)*this->cbLs;
	      this->shift_table[local_p][info_idx] |=(1<<(pm+2*mu));
            } else { 
	      this->shift_table[local_p][tab] = shift_cbsite*24*this->nsimd*sizeof(Float)*this->cbLs;
            }
	    /*
	     *   X X X A
	     * 
	     *   face_table[plus] = 4
	     *   shift_table[4,plus] = this->comm_offset[plus]
	     *
	     *   gather (face_table) -> buffer
             *   This should be copied to -> this->comm_offset[minus]
	     *   Reading the corresponding minus face from shift_table pulls A in the hopping term
	     */
	    bound_index[local_p][pm][mu] ++;
	  }


        } /*mu*/
       }/*x*/
      }/*y*/
     }/*z*/
    }/*t*/
  }/*pm*/


  if ( this->precon_5d ) { // Checkerboarding for 5d is different
    for(int s=0;s<this->simd_cbvol;s++){
      int idx=Nmu*NMinusPlus*2*s;
      for(int i=0;i<Nmu*NMinusPlus*2;i++){
	this->shift_table[1][idx+i]=this->shift_table[0][idx+i];
      }
      this->shift_table[1][idx+11] =1-this->shift_table[0][idx+11]; // Flip the parity
      if ( s==this->simd_cbvol-1 ) {
	this->shift_table[1][idx+8] =0;
      } else {
	this->shift_table[1][idx+8] = this->shift_table[1][idx+16];
      }
    }
  }

#if 0
      /*Loop naively over local lattice*/
  for ( t=0 ; t<lt ; t++ ) {
  for ( z=0 ; z<lz ; z++ ) {
  for ( y=0 ; y<ly ; y++ ) {
  for ( x=0 ; x<lx ; x++ ) {

    local_addr[3] = t;
    local_addr[2] = z;
    local_addr[1] = y;
    local_addr[0] = x;

    /* local_addr contains the coordinates of the point */
    /* we get the parity of the point                   */
    local_p = this->parity(local_addr);
    /* get the cbsite of the point for within parity */
    cbsite   = this->psite(local_addr,this->simd_latt);
    printf("%d %d %d %d Site %d\n",x,y,z,t,cbsite);
    for(int i=0;i<16;i++){
      printf("%d\t",this->shift_table[local_p][cbsite*16+i]);
    }
    printf("\n");

  }}}}
#endif
  //  int bound_index[Ncb][NMinusPlus][ND];
  for( cb = 0 ; cb < 2 ; cb++ ) {
  for( pm = 0 ; pm < 2 ; pm++ ) {
  for( mu = 0 ; mu < Nmu ; mu++ ) {
    if ( (!this->precon_5d) || cb == 0 ) {
      if ( bound_index[cb][pm][mu] != this->simd_nbound[mu] ) {
	printf("bfm<Float>::pointers Boundary size mismatch[%d][%d][mu=%d] : %d != %d \n",
	       cb,pm,mu,bound_index[cb][pm][mu],this->simd_nbound[mu]);
//	exit(-1);
      }
    }
  }
  }
  }

  return;
}

template <class Float>
int bfmdop<Float>::interleave_site(int mp,int mu, int site)
{
  return(mp+mu*NMinusPlus+site*NMinusPlus*ND*2);
}

//
// To switch to 5d preconditioning, I think I just need to
// make parity always return zero, and make psite not divide by two.
// introduce a flat precon_5d in the bfmarg class
//

template <class Float>
int bfmdop<Float>::parity(int x[4])
{
  if ( this->precon_5d ) return 0;
  else return((x[0]+x[1]+x[2]+x[3]) % 2);
}

template <class Float>
int bfmdop<Float>::psite(int addr[4],int latt[4])
{
  int site;
  site =  addr[0] + latt[0]*(addr[1]
                             + latt[1]*(addr[2]
                                        + latt[2]*addr[3]));
  if ( this->precon_5d ) return site;
  else return(site/2);
}

template <class Float>
void bfmdop<Float>::bgq_l1p_optimisation(int mode)
{
#define L1P_OPT
#ifdef L1P_OPT
#undef L1P_CFG_PF_USR
#define L1P_CFG_PF_USR  (0x3fde8000108ll)   /*  (64 bit reg, 23 bits wide, user/unpriv) */
  uint64_t cfg_pf_usr;
  if ( mode ) { 
    cfg_pf_usr =
        L1P_CFG_PF_USR_ifetch_depth(0)       
      | L1P_CFG_PF_USR_ifetch_max_footprint(1)   
      | L1P_CFG_PF_USR_pf_stream_est_on_dcbt 
      | L1P_CFG_PF_USR_pf_stream_establish_enable
      | L1P_CFG_PF_USR_pf_stream_optimistic
      | L1P_CFG_PF_USR_pf_adaptive_throttle(0xF) ;
    if ( sizeof(Float) == sizeof(double) ) {
      cfg_pf_usr |=  L1P_CFG_PF_USR_dfetch_depth(2)| L1P_CFG_PF_USR_dfetch_max_footprint(3)   ;
    } else {
      cfg_pf_usr |=  L1P_CFG_PF_USR_dfetch_depth(1)| L1P_CFG_PF_USR_dfetch_max_footprint(2)   ;
    }
  } else { 
    cfg_pf_usr = L1P_CFG_PF_USR_dfetch_depth(1)
      | L1P_CFG_PF_USR_dfetch_max_footprint(2)   
      | L1P_CFG_PF_USR_ifetch_depth(0)       
      | L1P_CFG_PF_USR_ifetch_max_footprint(1)   
      | L1P_CFG_PF_USR_pf_stream_est_on_dcbt 
      | L1P_CFG_PF_USR_pf_stream_establish_enable
      | L1P_CFG_PF_USR_pf_stream_optimistic
      | L1P_CFG_PF_USR_pf_stream_prefetch_enable;
  }

  *((uint64_t *)L1P_CFG_PF_USR) = cfg_pf_usr;
#endif

}

template <class Float>
void bfmdop<Float>::gather_proj(integer *sites,
			     Float     *psi,
			     integer comm_off, 
			     Float  *comm_buf,
			     integer nface,
			     int permute, 
			     int L,
			     int pm,
			     int mu)
{

  int me,thrvol,throff;

  this->thread_work(nface,me,thrvol,throff);

  integer args[8];
  args[0] = (integer) &sites[throff];
  args[1] = (integer) psi;
  args[2] = (integer) comm_off+throff;
  args[3] = (integer) comm_buf; // Fix me -- need offset into this
  args[4] = (integer) thrvol;
  args[5] = (integer) L;
  args[6] = pm+2*mu;
  args[7] = (integer)complex_i_simd;

  integer arg_p = (integer) args;

  if ( permute ) { 
    if ( sizeof(Float) == sizeof(double ) ) vmx_gather_proj_perm(arg_p);
    else                                    vmx_gather_proj_perm_s(arg_p);
  } else { 
    if ( sizeof(Float) == sizeof(double ) ) vmx_gather_proj(arg_p);
    else                                    vmx_gather_proj_s(arg_p);
  }  
  this->thread_barrier();
}

		   
template class bfmdop<float>;
template class bfmdop<double>;
