#include <chroma.h>
#include <actions/ferm/invert/syssolver_linop_cg_array.h>
#include <bfm.h>
#include <bfm_qdp.h>
#include <eigen/Krylov_5d.h>

typedef bfm bfm_t;
//typedef commQMPbagel1<double> bfm_t;

using namespace Chroma;

typedef LatticeFermion T;
typedef multi1d<LatticeFermion> T5;
typedef multi1d<LatticeColorMatrix> U;

//#define Printf if ( QMP_is_primary_node() ) printf
#define Printf printf

int main (int argc,char **argv )
{
  Chroma::initialize(&argc,&argv);


  /********************************************************
   * Command line parsing
   ********************************************************
   */
  if ( argc < 10 ) { 
   Printf("Usage: %s lx ly lz lt Ls Nwanted Tn off rad\n All must be even\n",argv[0]);
   Printf("Good test: D_oo 4 4 4 4 4 10 30 1.0 6.0");
   Printf("argc is %d\n",argc);
   for ( int i=0;i<argc;i++){
      Printf("%d %s\n",i,argv[i]);
   }
      exit(-1);
  }
  /********************************************************
   * Setup QDP
   ********************************************************
   */
  multi1d<int> nrow(Nd);
  nrow[0] = atoi(argv[1]);
  nrow[1] = atoi(argv[2]);
  nrow[2] = atoi(argv[3]);
  nrow[3] = atoi(argv[4]);
  int Ls = atoi(argv[5]);



  Layout::setLattSize(nrow);
  Layout::create();

  int lx = QDP::Layout::subgridLattSize()[0];
  int ly = QDP::Layout::subgridLattSize()[1];
  int lz = QDP::Layout::subgridLattSize()[2];
  int lt = QDP::Layout::subgridLattSize()[3];

  int threads = 64;
  bfmarg::Threads(threads);
  bfmarg::Reproduce(1);
  bfmarg::ReproduceChecksum(0);
  bfmarg::ReproduceMasterCheck(0);
  bfmarg::Verbose(0);

  /********************************************************
   * Setup DWF operator
   ********************************************************
   */
  bfmarg dwfa;
  bfm_t  dwf;
  dwfa.solver = DWFrb4d;
  dwfa.node_latt[0]  = lx;
  dwfa.node_latt[1]  = ly;
  dwfa.node_latt[2]  = lz;
  dwfa.node_latt[3]  = lt;
  dwfa.verbose=1;

  multi1d<int> procs = QDP::Layout::logicalSize();
  Printf("%d dim machine\n\t", procs.size());
  for(int mu=0;mu<4;mu++){
    Printf("%d ", procs[mu]);
    if ( procs[mu]>1 ) {
      dwfa.local_comm[mu] = 0;
    } else { 
      dwfa.local_comm[mu] = 1;
    }
  }
  Printf("\nLocal comm = ");
  for(int mu=0;mu<4;mu++){
    Printf("%d ", dwfa.local_comm[mu]);
  }
  Printf("\n");
  bfmarg::Threads(64);

  multi1d<int> ncoor = QDP::Layout::nodeCoord();

  Real M5(1.8);
  Real mq(0.0);
  dwfa.solver       = DWF; //DWFrb4d = 4d preconditioning. DWF = 5d preconditioning
  dwfa.precon_5d = 1;
  dwfa.Ls   = Ls;
  dwfa.M5   = toDouble(M5);
  dwfa.mass = toDouble(mq);
  dwfa.Csw  = 0.0;
  dwfa.max_iter = 1000;
  dwfa.residual = 1.e-10;
  Printf("Initialising bfm operator\n");

  bfm_qdp<double> dwf_qdp;
  dwf.init(dwfa);
  dwf_qdp.init(dwfa);

  /********************************************************
   * Gaussian gauge field
   ********************************************************
   */

  multi1d<LatticeColorMatrix>  u(Nd);
  for(int i=0;i<Nd; i++){
   u[i] = 1.0; 
  }
  HotSt(u);
  if(argc == 11) {ArchivGauge_t Header ; readArchiv(Header,u,argv[10]);}

  /********************************************************
   * Anti-Periodic Boundary conditions
   ********************************************************
   */

    multi1d<int> bcond (Nd);
    bcond[0] = bcond[1] = bcond[2] = 1; bcond[3] = -1;
    Handle<FermBC<T,U,U> > fbca (new SimpleFermBC<T,U,U>(bcond));
    fbca->modify(u);
    dwf.importGauge(u);
    dwf_qdp.importGauge(u);
  /********************************************************
   * Gaussian source and result vectors
   ********************************************************
   */
  multi1d<LatticeFermion> psi(Ls);
  multi1d<LatticeFermion> chi(Ls);
  multi1d<LatticeFermion> chi_qdp(Ls);

  for(int s=0;s<Ls;s++) gaussian(psi[s]);
  for(int s=0;s<Ls;s++) gaussian(chi[s]);
  for(int s=0;s<Ls;s++) gaussian(chi_qdp[s]);



  /********************************************************
   * Bagel internal single checkerboard vectors
   ********************************************************
   */
   Fermion_t psi_h[2];
   psi_h[0] = dwf.allocFermion();
   psi_h[1] = dwf.allocFermion();
   Fermion_t chi_h[2];
   chi_h[0] = dwf.allocFermion();
   chi_h[1] = dwf.allocFermion();

  /********************************************************
   * Import gauge field to BAGEL
   ********************************************************
   */
  //dwf.importGauge(u);
  //dwf.inner(psi_h[0], chi_h[0]);
#undef DEBUG

  /*Import this checkerboard of source field to bagel*/
  Printf("Importing psi field cb %d\n",1);
  dwf.importFermion(psi,psi_h[0],0);
  dwf.importFermion(psi,psi_h[1],1);
 
  // Fill the other checkerboard of result with noise
  for(int s=0;s<Ls;s++) gaussian(chi[s]);
  for(int s=0;s<Ls;s++) chi_qdp[s] = chi[s];
  dwf.importFermion(chi,chi_h[0],0);
  dwf.importFermion(chi,chi_h[1],1);

#define NITER 1
  for(int i=0;i<NITER;i++) {
    //    dwf.axpy(chi_h[0],psi_h[0],psi_h[0],0.0)  ;
    //    dwf.axpy(chi_h[1],psi_h[0],psi_h[0],0.0)  ;
     dwf.CGNE(chi_h,psi_h);
  }
    /*#define Ncg 1
    Real residuals[Ncg] ; 
    int  max_iter[Ncg];
    for(int i=0;i<Ncg;i++) residuals[i] = 1.0e-12;
    for(int i=0;i<Ncg;i++) max_iter[i]  = 1000;
    dwf_restarted_invert<float,double>(CG_PREC_M,
				       chi, 
				       psi,
				       u,
				       "pooh.xml",
				       Ls,
				       mq,
				       M5,Ncg,residuals,max_iter);*/

  dwf.exportFermion(chi,chi_h[0],0);
  dwf.exportFermion(chi,chi_h[1],1);
 
#define CHECKIT
#ifdef  CHECKIT 
  /********************************************************
   * QDP Linop
   ********************************************************
   */
   multi1d<int> bcs(Nd);
   bcs[0] = bcs[1] = bcs[2] = bcs[3] = 1;

   Handle< FermBC<T,U,U> > fbc(new SimpleFermBC< T, U, U >(bcs));
   Handle<CreateFermState<T,U,U> > cfs( new CreateSimpleFermState<T,U,U>(fbc));
   EvenOddPrecDWFermActArray  S_f(cfs, M5, mq, Ls);
   Handle< FermState<T,U,U> > fs( S_f.createState(u) );

   Handle< EvenOddPrecLinearOperatorArray<T,U,U> > MM(S_f.precLinOp(fs,mq));
   Handle< LinearOperatorArray<T> > HM(S_f.precLinOp(fs,mq));
  // Check the result
  SysSolverCGParams invParam;
  invParam.RsdCG        = 1.0e-8;
  invParam.RsdCGRestart = 1.0e-8;
  invParam.MaxCG        = dwfa.max_iter ;
  invParam.MaxCGRestart = dwfa.max_iter ;

  Handle<LinOpSystemSolverArray<T> > Hsolver = 
    new LinOpSysSolverCGArray<T>(HM,invParam);
  Handle<PrecFermAct5DQprop<T,U,U> > solver =
    new PrecFermAct5DQprop<T,U,U>(MM,Hsolver);
  SystemSolverResults_t  stats =  (*solver)(chi_qdp,psi);
  int n_count = stats.n_count;

  Double n2 = 0.0;
  for(int s=0;s<Ls;s++) { 
    Double sn2 = norm2(chi[s]-chi_qdp[s]);
    n2+=sn2;
#ifdef DEBUG
    QDPIO::cout << "|| Bagel - QDP || ["<<s<<"] = "<< sn2 << endl;
#endif
  }
  QDPIO::cout << "|| Bagel - QDP || = "<< n2 << endl;
  
  n2 = 0.0;
  for(int s=0;s<Ls;s++) { 
    Double sn2 = norm2(chi[s]);
    n2+=sn2;
#ifdef DEBUG
    QDPIO::cout << "|| Bagel || ["<<s<<"] = "<< sn2 << endl;
#endif
  }
  QDPIO::cout << "|| Bagel || = "<< n2 << endl;
  
  n2 = 0.0;
  for(int s=0;s<Ls;s++) {
    Double sn2 = norm2(chi_qdp[s]);
    n2+=sn2;
#ifdef DEBUG
    QDPIO::cout << "|| QDP || ["<<s<<"] = "<< sn2 << endl;
#endif
  }
  QDPIO::cout << "|| QDP || = "<< n2 << endl;
#endif


  /********************************************************
   * EigenCalculation: Calculate K eigenvectors
   ********************************************************
   */
   int K = atoi(argv[6]);
   int M = K + 20;
   int Tn = atoi(argv[7]);

   multi1d<multi1d<LatticeFermion> > Eigs(0); 
   multi1d<Complex> vals(0);
   
	QDPIO::cout << "Set up Lanczos 5d..." << endl;
  	Lanczos_5d<double> eig(dwf_qdp);		///Constructor
	///Eigensolver parameters	
	eig.u = u;				///Read gauge field
	eig.M = M;				///Nsearch > Nwanted
	eig.K = K;				///Nwanted
	eig.conv = 1e-8;			///allowed residual
	eig.D = DDAGD;	 			///Type of multiplication;
	eig.Tn = Tn;				///Order of Chebyshev
	eig.ch_off = atof(argv[8]);		///> max wanted eigenvalue of G5D 
	eig.ch_rad = atof(argv[9]);		///> spectral radius of G5D 
	eig.sh = false; eig.ch_shift = 0.0;	///shift to get internal eigenvalues
	eig.prec = 1;				///preconditioned : 1 or not : 0
	QDPIO::cout << "Call eigensolver..." << endl;
	eig.Run();

  /********************************************************
   * Deflate with calculated evecs
   ********************************************************
   */
	multi1d<LatticeFermion> chi_deflated(Ls);
	for(int s=0;s<Ls;s++) gaussian(chi_deflated[s]);
  	for(int s=0;s<Ls;s++) chi_deflated[s] = chi[s];

  	dwf.importFermion(psi,psi_h[0],0);
  	dwf.importFermion(psi,psi_h[1],1);
  	dwf.importFermion(chi_deflated,chi_h[0],0);
  	dwf.importFermion(chi_deflated,chi_h[1],1);

  	dwf.CGNE(chi_h,psi_h,eig.bq,eig.bl);

  	dwf.exportFermion(chi_deflated,chi_h[0],0);
  	dwf.exportFermion(chi_deflated,chi_h[1],1);

	Real diff = 0.0;
	for(int s=0;s<Ls;s++) { diff = norm2(chi[s] - chi_deflated[s]); }
	QDPIO::cout << "|chi_bagel - chi_deflated| = " << sqrt(diff) << endl;

	diff = 0.0;
	for(int s=0;s<Ls;s++) { diff = norm2(chi_qdp[s] - chi_deflated[s]); }
	QDPIO::cout << "|chi_qdp - chi_deflated| = " << sqrt(diff) << endl;
	Printf("Done\n");

}








